module Apriori where

import qualified Data.Set as Set
import Control.Monad.Par (parMap, runPar)
import Control.DeepSeq
import Control.Parallel.Strategies (using, parListChunk, rdeepseq)

-- definitions
type Support = Double

data Itemset = Itemset (Set.Set String) deriving (Eq, Ord)
instance Show Itemset where
  show (Itemset i) = show $ Set.toList i
instance NFData Itemset where
  rnf (Itemset i) = rnf i

-- calculate support
supportCount :: [Itemset] -> Itemset -> Int
supportCount transactions (Itemset i) =
  length $ filter (Set.isSubsetOf i) $ map (\(Itemset x) -> x) transactions

support :: [Itemset] -> Itemset -> Support
support transactions is =
  fromIntegral (supportCount transactions is) / fromIntegral (length transactions)

-- util
deduplicate :: Ord a => [a] -> [a]
deduplicate l = Set.toList $ Set.fromList l

allSubsets :: Set.Set a -> [Set.Set a]
allSubsets s = Set.toList $ Set.powerSet s

allMaximalProperSubsets :: Set.Set a -> [Set.Set a]
allMaximalProperSubsets s = filter (\x -> (Set.size x) == (Set.size s) - 1) $ allSubsets s

-- apriori
firstCandItemsets :: [Itemset] -> [Itemset]
firstCandItemsets transactions =
  deduplicate $ concatMap (\(Itemset is) -> map (Itemset . Set.singleton) $ Set.toList is) transactions

firstFreqItemsets :: Support -> [Itemset] -> [Itemset]
firstFreqItemsets minSupp transactions =
  filter (\candIs -> support transactions candIs > minSupp) (firstCandItemsets transactions)

nextCandItemsets :: [Itemset] -> [Itemset]
nextCandItemsets itemsets =
  -- two k-1 sets should differ in exactly 1 element before joining
  let validate a b = (Set.size $ a `Set.difference` b) == 1 in
  -- self join
  let nextIss = [Itemset (a `Set.union` b) | (Itemset a) <- itemsets, (Itemset b) <- itemsets, validate a b] in
  -- prune
  let subsetsFrequent (Itemset is) = all (\s -> (Itemset s) `elem` itemsets) (allMaximalProperSubsets is) in
  deduplicate $ filter subsetsFrequent nextIss

nextFreqItemsets :: Support -> [Itemset] -> [Itemset] -> Maybe ([Itemset], [Itemset])
nextFreqItemsets _ _ [] = Nothing
nextFreqItemsets minSupp transactions curr =
  Just (curr, next)
    where
      next = filter (\candIs -> support transactions candIs > minSupp) (nextCandItemsets curr)

-- parallel pkg
parFirstFreqItemsets :: Int -> Support -> [Itemset] -> [Itemset]
parFirstFreqItemsets n minSupp transactions =
  let fullFirstIss = map (\candIs -> (candIs, support transactions candIs > minSupp)) (firstCandItemsets transactions) `using` parListChunk n rdeepseq in
  concat $ [[candIs] | (candIs, isGood) <- fullFirstIss, isGood]

parNextCandItemsets :: Int -> [Itemset] -> [Itemset]
parNextCandItemsets n itemsets =
  -- two k-1 sets should differ in exactly 1 element before joining
  let validate a b = (Set.size $ a `Set.difference` b) == 1 in
  -- self join
  let nextIss = [Itemset (a `Set.union` b) | (Itemset a) <- itemsets, (Itemset b) <- itemsets, validate a b] in
  -- prune
  let subsetsFrequent (Itemset is) = all (\s -> (Itemset s) `elem` itemsets) (allMaximalProperSubsets is) in
  -- get full next itemsets [(is, good)] in parallel
  let fullNextIss = map (\nextIs -> (nextIs, subsetsFrequent nextIs)) nextIss `using` parListChunk n rdeepseq in
  deduplicate $ concat $ [[nextIs] | (nextIs, isGood) <- fullNextIss, isGood]

parNextFreqItemsets :: Int -> Support -> [Itemset] -> [Itemset] -> Maybe ([Itemset], [Itemset])
parNextFreqItemsets _ _ _ [] = Nothing
parNextFreqItemsets n minSupp transactions curr =
  let fullNextCandIss = map (\candIs -> (candIs, support transactions candIs > minSupp)) (parNextCandItemsets n curr) `using` parListChunk n rdeepseq in
  let next = concat $ [[candIs] | (candIs, isGood) <- fullNextCandIss, isGood] in
  Just (curr, next)

-- monad-parallel
monadParFirstFreqItemsets :: Support -> [Itemset] -> [Itemset]
monadParFirstFreqItemsets minSupp transactions =
  let fullFirstIss = runPar $ parMap (\candIs -> (candIs, support transactions candIs > minSupp)) (firstCandItemsets transactions) in
  concat $ [[candIs] | (candIs, isGood) <- fullFirstIss, isGood]

monadParNextCandItemsets :: [Itemset] -> [Itemset]
monadParNextCandItemsets itemsets =
  -- two k-1 sets should differ in exactly 1 element before joining
  let validate a b = (Set.size $ a `Set.difference` b) == 1 in
  -- self join
  let nextIss = [Itemset (a `Set.union` b) | (Itemset a) <- itemsets, (Itemset b) <- itemsets, validate a b] in
  -- prune
  let subsetsFrequent (Itemset is) = all (\s -> (Itemset s) `elem` itemsets) (allMaximalProperSubsets is) in
  -- get full next itemsets [(is, good)] in parallel
  let fullNextIss = runPar $ parMap (\nextIs -> (nextIs, subsetsFrequent nextIs)) nextIss in
  deduplicate $ concat $ [[nextIs] | (nextIs, isGood) <- fullNextIss, isGood]

monadParNextFreqItemsets :: Support -> [Itemset] -> [Itemset] -> Maybe ([Itemset], [Itemset])
monadParNextFreqItemsets _ _ [] = Nothing
monadParNextFreqItemsets minSupp transactions curr =
  let fullNextCandIss = runPar $ parMap (\candIs -> (candIs, support transactions candIs > minSupp)) (monadParNextCandItemsets curr) in
  let next = concat $ [[candIs] | (candIs, isGood) <- fullNextCandIss, isGood] in 
  Just (curr, next)
